module RKelly
  module Visitors
    class ParentBuilder < Visitor

      def build node
        link_parents node, nil
        node
      end

      ## define methods for traverse different kinds of nodes
      ## consts are from RKelly::Visitors::Visitor

      # dont need to link the value for termianl nodes
      TERMINAL_NODES.each do |type|
        define_method(:"visit_#{type}Node") { |o| o.value }
      end

      BINARY_NODES.each do |type|
        define_method(:"visit_#{type}Node") do |o|
          [o.left && link_parents(o.left, o), o.value && link_parents(o.value, o)]
        end
      end

      ARRAY_VALUE_NODES.each do |type|
        define_method(:"visit_#{type}Node") do |o|
          o.value && o.value.map { |v| v ? link_parents(v, o) : nil }
        end
      end

      NAME_VALUE_NODES.each do |type|
        define_method(:"visit_#{type}Node") do |o|
          [o.name.to_s.to_sym, o.value ? link_parents(o.value, o) : nil]
        end
      end

      SINGLE_VALUE_NODES.each do |type|
        define_method(:"visit_#{type}Node") do |o|
          link_parents(o.value, o) if o.value
        end
      end

      PREFIX_POSTFIX_NODES.each do |type|
        define_method(:"visit_#{type}Node") do |o|
          link_parents(o.operand, o)
        end

      end

      CONDITIONAL_NODES.each do |type|
        define_method(:"visit_#{type}Node") do |o|
          [ link_parents(o.conditions, o),
            link_parents(o.value, o),
            o.else ? link_parents(o.else, o) : nil
          ]
        end
      end
      FUNC_CALL_NODES.each do |type|
        define_method(:"visit_#{type}Node") do |o|
          [link_parents(o.value, o), link_parents(o.arguments, o)]
        end
      end
      FUNC_DECL_NODES.each do |type|
        define_method(:"visit_#{type}Node") do |o|
          [
            o.value ? o.value : nil,
            o.arguments.map { |x| link_parents(x, o) },
            link_parents(o.function_body, o)
          ]
        end
      end

      def visit_ForNode(o)
        [
          o.init ? link_parents(o.init, o) : nil,
          o.test ? link_parents(o.test, o) : nil,
          o.counter ? link_parents(o.counter, o) : nil,
          link_parents(o.value, o)
        ]
      end

      def visit_ForInNode(o)
        [
          link_parents(o.left, o),
          link_parents(o.right, o),
          link_parents(o.value, o)
        ]
      end

      def visit_TryNode(o)
        [
          link_parents(o.value, o),
          o.catch_var ? o.catch_var : nil,
          o.catch_block ? link_parents(o.catch_block, o) : nil,
          o.finally_block ? link_parents(o.finally_block, o) : nil
        ]
      end

      def visit_BracketAccessorNode(o)
        [
          link_parents(o.value, o),
          link_parents(o.accessor, o)
        ]
      end

      def visit_DotAccessorNode(o)
        link_parents(o.value, o)
      end

      private
      def link_parents node, parent
        node.parent = parent

        # call the visit method to traverse potential subnodes depends on the node class
        # error if the node is not in the predefined list
        #
        raise "No visit method for '#{node.class}'" unless self.respond_to?("visit_#{node.class.name.split(/::/)[-1]}")
        self.send(:"visit_#{node.class.name.split(/::/)[-1]}", node)
      end
    end
  end
end
